"""itkpix-efuse-codec CLI"""

import click
import sys

import itkpix_efuse_codec
from itkpix_efuse_codec import EfuseData

@click.group(name="codec")
def cli():
    """The efuse codec CLI group."""

@cli.command()
@click.pass_context
@click.argument("probe-location", type=str)
@click.argument("chip-sn", type=str)
def encode(ctx, probe_location, chip_sn):
    """
    Construct the encoded E-fuse data
    """

    chip_sn = int(chip_sn, 16)
    encoded_binary_string = itkpix_efuse_codec.encode(probe_location, chip_sn)
    encoded_int = int(encoded_binary_string, 2)

    if ctx.obj["VERBOSE"] :
        print(f"Input:")
        print(f"   probe_location name  : {probe_location}")
        print(f"   chip serial number   : {chip_sn} (bin: {bin(chip_sn)}, hex: {hex(chip_sn)})")
        print(f"Encoded:")
        print(f"   encoded e-fuse (bin) : {encoded_binary_string} ({len(encoded_binary_string)} bits)")
        print(f"                  (hex) : 0x{hex(encoded_int)[2:]}")
    else :
        print(f"0x{hex(encoded_int)[2:]}")

@cli.command()
@click.pass_context
@click.argument("efuse-data", type = str)
def decode(ctx, efuse_data) :
    """
    Decode a retrieved e-fuse data word.
    """

    
    efuse_data = int(efuse_data, 16)
    decoded_binary_string = itkpix_efuse_codec.decode(efuse_data)
    decoded_int = int(decoded_binary_string, 2)

    if ctx.obj["VERBOSE"] :
        efuse_data = EfuseData(decoded_binary_string)
        print(f"Input:")
        print(f"    e-fuse data   (bin): {bin(decoded_int)}")
        print(f"                  (hex): {hex(decoded_int)}")
        print(f"Decoded:")
        print(f"    chip serial number : {hex(efuse_data.chip_sn())}")
        print(f"    probe location     : {efuse_data.probe_location_name()} (id: {hex(efuse_data.probe_location_id())})")
    else :
        print(f"0x{hex(decoded_int)[2:]}")



import sys
import pathlib

import os
import subprocess
from setuptools.command.sdist import sdist
from setuptools.command.install import install

# the directory containing this file
HERE = pathlib.Path(__file__).parent

# the text of the README file
README = (HERE / "README.md").read_text()

__version__ = "0.2.0"


try:
    from skbuild import setup
except ImportError:
    print(
        "Please update pip, you need pip 10 or greater,\n"
        " or you need to install the PEP 518 requirements in pyproject.toml yourself",
        file=sys.stderr,
    )
    raise


def check_submodules():
    """verify that the submodules are checked out and clean
    use `git submodule update --init`; on failure
    """
    if not os.path.exists(".git"):
        return
    with open(".gitmodules") as f:
        for line in f:
            if "path" in line:
                p = line.split("=")[-1].strip()
                if not os.path.exists(p):
                    raise ValueError("Submodule {} missing".format(p))

    proc = subprocess.Popen(["git", "submodule", "status"], stdout=subprocess.PIPE)
    status, _ = proc.communicate()
    status = status.decode("ascii", "replace")
    for line in status.splitlines():
        if line.startswith("-") or line.startswith("+"):
            raise ValueError("Submodule not clean: {}".format(line))


class SdistChecker(sdist):
    """ check submodules on sdist to prevent incomplete tarballs """

    def run(self):
        check_submodules()
        sdist.run(self)


class InstallChecker(install):
    """ check submodules on install to prevent failed installs """

    def run(self):
        check_submodules()
        install.run(self)


cmake_compiler_defines = f"-DBUILD_PYTHON=on -DVERSION_INFO={__version__}"
print(f"cmake compiler defines = {cmake_compiler_defines}")

setup(
    name="hamming_codec",
    version=__version__,
    description="Simple encode/decode utilities for bit-error correcting Hamming codes",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/dantrim/hamming-codec",
    author="Daniel Joseph Antrim",
    author_email="dantrim1023@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    package_dir={"": "src/python"},
    packages=["hamming_codec", "cli"],
    install_requires=["click"],
    cmake_args=["-DBUILD_PYTHON=on"],
    cmake_install_dir="src/python/hamming_codec",
    entry_points={"console_scripts": ["hamming=cli:cli.hamming"]},
    cmdclass={"install": InstallChecker, "sdist": SdistChecker},
    include_package_data=True,
)



from ._hamming_codec import (  # noqa: F401
    encode,
    decode,
    ParityLocation,
    # n_parity_bits,
    # compute_parity_bit_positions,
    # compute_parity_bits,
    # n_parity_bits_required,
    # int2bin,
)
from ._hamming_codec import __version__  # noqa: F401



from functools import reduce
import operator as op


def n_parity_bits_required(n_bits: int) -> int:

    p = 1
    while True:
        lhs = 2 ** p
        rhs = p + n_bits + 1
        if lhs >= rhs:
            break
        p += 1
    return p


def compute_parity_bits(binary_string: str, positions: list, inclusive: bool) -> list:

    parity_bits = [0 for _ in positions]

    for i, p in enumerate(positions):
        mask = 1 << i
        filter_func = {
            False: lambda d: (mask & d != 0) and (mask != d),
            True: lambda d: (mask & d != 0),
        }[inclusive]
        r_pos = [x for x in list(filter(filter_func, range(len(binary_string) + 1)))]
        data_sel = [int(list(binary_string)[d - 1], 2) for d in r_pos]
        xor = reduce(op.xor, data_sel)
        if xor == 1:
            parity_bits[i] = 1
    return parity_bits


def encode(data: int, n_bits: int) -> str:

    binary_string = bin(data)[2:]

    if n_bits < len(binary_string):
        print(
            f"ERROR: Requested data size ({n_bits} bits) is smaller than the binary representation of the input data (={data})"
        )
        return -1
    # pad
    binary_string = f"{binary_string:0>{n_bits}}"

    # parity bits are at powers of 2
    n_parity_bits = n_parity_bits_required(n_bits)
    parity_bit_positions = [2 ** i - 1 for i in range(n_parity_bits)]

    len_encoded_message = n_parity_bits + n_bits
    parity_bit_positions = list(
        filter(lambda x: x < len_encoded_message, parity_bit_positions)
    )

    binary_string_reversed = "".join(reversed(binary_string))

    # placeholder string
    seed_string = "".join(["x" for _ in range(len_encoded_message)])
    seed_string = list(seed_string)

    data_idx = 0
    for idx in range(len_encoded_message):
        if idx not in parity_bit_positions:
            seed_string[idx] = list(binary_string_reversed)[data_idx]
            data_idx += 1
    seed_string = "".join(seed_string)

    # compute the values for the parity bits
    parity_bits = compute_parity_bits(seed_string, parity_bit_positions, False)

    # emplace the values of the parity bits in the flagged positions
    parity_bit_idx = 0
    encoded_string = list(seed_string)
    for i, v in enumerate(encoded_string):
        if v.lower() == "x":
            encoded_string[i] = parity_bits[parity_bit_idx]
            parity_bit_idx += 1
    encoded_string = "".join(map(str, reversed(encoded_string)))
    return encoded_string


def decode(binary_string: str, n_bits: int) -> str:

    # binary string must not have the "0b" preceding characters

    n_parity_bits = n_parity_bits_required(n_bits)
    parity_bit_positions = [2 ** i - 1 for i in range(n_parity_bits)]
    parity_bit_positions = list(
        filter(lambda x: x < len(binary_string), parity_bit_positions)
    )

    binary_string_reversed = "".join(reversed(list(binary_string)))
    parity_bits = compute_parity_bits(
        binary_string_reversed, parity_bit_positions, True
    )
    error_position = int("".join(reversed(list(map(str, parity_bits)))), 2)
    decoded_string = list(binary_string_reversed)
    if error_position > 0:
        # flip the bit at the index where the error is located
        decoded_string[error_position - 1] = {"0": "1", "1": "0"}[
            decoded_string[error_position - 1]
        ]

    # remove the parity bits from the decoded string to get the message data
    decoded_string = [
        v for i, v in enumerate(decoded_string) if i not in parity_bit_positions
    ]
    decoded_string = "".join(reversed(decoded_string))
    return decoded_string



"""hamming CLI"""

import click
import sys

import hamming_codec


@click.group(name="codec")
def cli():
    """The codec CLI group."""


@cli.command()
@click.pass_context
@click.argument("input", type=str)
@click.argument("n-bits", type=int)
@click.argument("parity-location", type=str, default="DEFAULT")
def encode(ctx, input, n_bits, parity_location):
    """
    Encode the provided input data word, which is interpreted
    as being word of the specified number of bits.
    """
    # decode the input string as a hexadecimal string
    input_data = int(input, 16)
    n_bits = int(n_bits)
    if n_bits < 4:
        raise ValueError("Cannot encode values that are less than 4 bits in length!")
    # convert to binary string of `n_bits` bits
    input_data_binary_string = f"{bin(input_data)[2:]:0>{n_bits}}"
    parity_location_map = {
        "DEFAULT": hamming_codec.ParityLocation.DEFAULT,
        "MSB": hamming_codec.ParityLocation.MSB,
        "LSB": hamming_codec.ParityLocation.LSB,
    }
    if parity_location not in parity_location_map:
        raise ValueError(f'Invalid parity location provided: "{parity_location}"')
    parity_location = parity_location_map[parity_location]

    # encode
    encoded_binary_string = hamming_codec.encode(input_data, n_bits, parity_location)
    encoded_int = int(encoded_binary_string, 2)
    if ctx.obj["VERBOSE"]:
        print(f"Input value         : 0x{input}, size = {n_bits} bits")
        print(f"Input value (bin)   : 0b{input_data_binary_string}")
        print(f"Encoded value       : 0x{hex(encoded_int)[2:]}")
        print(
            f"Encoded value (bin) : 0b{encoded_binary_string}, size = {len(encoded_binary_string)} bits"
        )
    else:
        print(f"0x{hex(encoded_int)[2:]} {len(encoded_binary_string)}", file=sys.stdout)


@cli.command()
@click.pass_context
@click.argument("input", type=str)
@click.argument("n-bits", type=int)
@click.argument("parity-location", type=str, default="DEFAULT")
@click.argument("n-parity-bits", type=int, default=0)
def decode(ctx, input, n_bits, parity_location, n_parity_bits):
    """
    Decode the input message that is the specified number of bits in
    length.
    """
    # input string as a hexadecimal string
    input_data = int(input, 16)
    n_bits_input = int(n_bits)
    if n_bits_input < 4:
        raise ValueError("Cannot decode values that are less than 4 bits in length!")

    # convert to binary string of `n_bits_input` bits
    input_data_binary_string = f"{bin(input_data)[2:]:0>{n_bits_input}}"

    parity_location_map = {
        "DEFAULT": hamming_codec.ParityLocation.DEFAULT,
        "MSB": hamming_codec.ParityLocation.MSB,
        "LSB": hamming_codec.ParityLocation.LSB,
    }
    if parity_location not in parity_location_map:
        raise ValueError(f'Invalid parity location provided: "{parity_location}"')
    parity_location = parity_location_map[parity_location]

    # decode
    decoded_binary_string = hamming_codec.decode(
        input_data, n_bits_input, parity_location, n_parity_bits
    )
    decoded_int = int(decoded_binary_string, 2)
    if ctx.obj["VERBOSE"]:
        print(f"Input value         : 0x{input}, size = {n_bits_input} bits")
        print(f"Input value (bin)   : 0b{input_data_binary_string}")
        print(f"Decoded value       : 0x{hex(decoded_int)[2:]}")
        print(
            f"Decoded value (bin) : 0b{decoded_binary_string}, size = {len(decoded_binary_string)} bits"
        )
    else:
        print(f"0x{hex(decoded_int)[2:]} {len(decoded_binary_string)}", file=sys.stdout)



"""hamming CLI"""

import click

from . import codec_cli


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("-v", "--verbose", is_flag=True, default=False)
@click.pass_context
def hamming(ctx, verbose):
    """Top-level entrypoint into hamming-codec utilities."""

    # ensure that ctx.obj exists and is a dict
    ctx.ensure_object(dict)
    # pass the verbose flag to the sub-commands
    ctx.obj["VERBOSE"] = verbose


hamming.add_command(codec_cli.encode)
hamming.add_command(codec_cli.decode)



"""itkpix-efuse-codec CLI"""

import click

from . import codec_cli

@click.group(context_settings = dict(help_option_names=["-h", "--help"]))
@click.option("-v", "--verbose", is_flag=True, default=False)
@click.pass_context
def efuse(ctx, verbose):
    """Top-level entrypoint into itkpix-efuse-codec utilities."""

    # ensure that ctx.obj exist and is dict
    ctx.ensure_object(dict)

    # pass the verbose flag to sub-commands
    ctx.obj["VERBOSE"] = verbose

efuse.add_command(codec_cli.encode)
efuse.add_command(codec_cli.decode)
